"""Serializable event-related classes."""

from __future__ import annotations

from datetime import datetime
import enum
import logging
from typing import Any, Dict, Hashable

from trulens.core.utils import serial as serial_utils
from typing_extensions import TypedDict

logger = logging.getLogger(__name__)


class EventRecordType(enum.Enum):
    """The enumeration of the possible record types for an event."""

    SPAN = "SPAN"


class Trace(TypedDict):
    """The type hint for a trace dictionary."""

    trace_id: str
    parent_id: str
    span_id: str


class Event(serial_utils.SerialModel, Hashable):
    """The class that represents a single event data entry."""

    event_id: str
    """
    The unique identifier for the event.
    """

    record: Dict[str, Any]
    """
    For a span, this is an object that includes:
    - name: the function/procedure that emitted the data
    - kind: SPAN_KIND_TRULENS
    - parent_span_id: the unique identifier for the parent span
    - status: STATUS_CODE_ERROR when the span corresponds to an unhandled exception. Otherwise, STATUS_CODE_UNSET.
    """

    record_attributes: Dict[str, Any]
    """
    Attributes of the record that can either come from the user, or based on the TruLens semantic conventions.
    """

    record_type: EventRecordType
    """
    Specifies the kind of record specified by this row. This will always be "SPAN" for TruLens.
    """

    resource_attributes: Dict[str, Any]
    """
    Reserved.
    """

    start_timestamp: datetime
    """
    The timestamp when the span started. This is a UNIX timestamp in milliseconds.
    Note: The Snowflake event table uses the TIMESTAMP_NTZ data type for this column.
    """

    timestamp: datetime
    """
    The timestamp when the span concluded. This is a UNIX timestamp in milliseconds.
    Note: The Snowflake event table uses the TIMESTAMP_NTZ data type for this column.
    """

    trace: Trace
    """
    The trace context information for the span.
    """

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    def __hash__(self):
        return self.trace["span_id"]



"""Serializable groundtruth-related classes."""

from __future__ import annotations

import logging
from typing import Dict, Hashable, Optional, Sequence

from pydantic import BaseModel
from pydantic import Field
from pydantic import ValidationError
from trulens.core.schema import types as types_schema
from trulens.core.utils import json as json_utils
from trulens.core.utils import serial as serial_utils

logger = logging.getLogger(__name__)


class VirtualGroundTruthSchemaMapping(BaseModel):
    query: str = Field(
        ...,
        description="Column name in the user's table mapping to the 'query' field.",
    )
    query_id: str = Field(
        ...,
        description="Column name in the user's table mapping to the 'query_id' field.",
    )
    expected_response: Optional[str] = Field(
        None, description="Column name for the 'expected_response' field."
    )
    expected_chunks: Optional[str] = Field(
        None, description="Column name for the 'expected_chunks' field."
    )
    dataset_id: Optional[str] = Field(
        None,
        description="Column name for the 'dataset_id' field.",  # TODO: is this still relevant for virtual?
    )

    @classmethod
    def validate_mapping(
        cls, schema_mapping: Dict[str, str]
    ) -> "VirtualGroundTruthSchemaMapping":
        """
        Validate and parse the schema mapping dictionary.

        Args:
            schema_mapping (Dict[str, str]): User-provided schema mapping.

        Returns:
            SchemaMapping: Parsed and validated schema mapping.

        Raises:
            ValidationError: If mandatory fields are missing or invalid.
        """
        try:
            return cls(**schema_mapping)
        except ValidationError as e:
            raise ValueError(f"Invalid schema mapping: {e}")


class GroundTruth(serial_utils.SerialModel, Hashable):
    """The class that represents a single ground truth data entry."""

    ground_truth_id: types_schema.GroundTruthID  # str
    """The unique identifier for the ground truth."""

    query: str
    """The query for which the ground truth is provided."""

    query_id: Optional[str] = None
    """Unique identifier for the query."""

    expected_response: Optional[str] = (
        None  # expected response can be empty in GT datasets
    )
    """The expected response for the query."""

    expected_chunks: Optional[Sequence[Dict]] = None
    """Expected chunks for the ground truth."""

    meta: Optional[types_schema.Metadata] = (
        None  # TODO: which naming are we exactly going with - meta vs metadata?
    )
    """Metadata for the ground truth."""

    dataset_id: types_schema.DatasetID  # str
    """The dataset ID to which this ground truth belongs.
    See [Dataset.dataset_id][trulens.core.schema.dataset.Dataset.dataset_id]."""

    def __init__(
        self,
        dataset_id: types_schema.DatasetID,
        query: str,
        query_id: Optional[str] = None,
        expected_response: Optional[str] = None,
        expected_chunks: Optional[Sequence[Dict]] = None,
        meta: Optional[types_schema.Metadata] = None,
        ground_truth_id: Optional[types_schema.GroundTruthID] = None,
        **kwargs,
    ):
        kwargs["query"] = query
        kwargs["query_id"] = query_id
        kwargs["dataset_id"] = dataset_id
        kwargs["expected_response"] = expected_response
        kwargs["expected_chunks"] = expected_chunks
        kwargs["meta"] = meta if meta is not None else {}

        super().__init__(
            ground_truth_id="temporary", **kwargs
        )  # will be updated below

        if ground_truth_id is None:
            ground_truth_id = json_utils.obj_id_of_obj(
                json_utils.jsonify(self), prefix="ground_truth"
            )
        self.ground_truth_id = ground_truth_id

    def __hash__(self):
        return hash(self.ground_truth_id)


# HACK013: Need these if using __future__.annotations .
GroundTruth.model_rebuild()



"""
# Synchronization/Async Utilities

NOTE: we cannot name a module "async" as it is a Python keyword.

## Synchronous vs. Asynchronous

Some functions in TruLens come with asynchronous versions. Those use "async
def" instead of "def" and typically start with the letter "a" in their name with
the rest matching their synchronous version.

Due to how Python handles such functions and how they are executed, it is
relatively difficult to reshare code between the two versions. Asynchronous
functions are executed by an async loop (see
[EventLoop](https://docs.python.org/3/library/asyncio-eventloop.html)). Python
prevents any threads from having more than one running loop meaning one may not
be able to create one to run some async code if one has already been
created/running in the thread. The method `sync` here, used to convert an async
computation into a sync computation, needs to create a new thread. The impact of
this, whether overhead, or record info, is uncertain.

### What should be Sync/Async?

Try to have all internals be async but for users we may expose sync versions via
the `sync` method. If internals are async and don't need exposure, don't need to
provide a synced version.

"""

import asyncio
import inspect
import logging
from threading import current_thread
from typing import Awaitable, Callable, TypeVar, Union

from trulens.core.utils import python as python_utils
from trulens.core.utils import threading as threading_utils

logger = logging.getLogger(__name__)

T = TypeVar("T")
A = TypeVar("A")
B = TypeVar("B")

MaybeAwaitable = Union[T, Awaitable[T]]
"""Awaitable or not.

May be checked with [isawaitable][inspect.isawaitable].
"""

CallableMaybeAwaitable = Union[Callable[[A], B], Callable[[A], Awaitable[B]]]
"""Function or coroutine function.

May be checked with
[is_really_coroutinefunction][trulens.core.utils.python.is_really_coroutinefunction].
"""

CallableAwaitable = Callable[[A], Awaitable[B]]
"""Function that produces an awaitable / coroutine function."""

ThunkMaybeAwaitable = Union[
    python_utils.Thunk[T], python_utils.Thunk[Awaitable[T]]
]
"""Thunk or coroutine thunk.

May be checked with
[is_really_coroutinefunction][trulens.core.utils.python.is_really_coroutinefunction].
"""


async def desync(
    func: CallableMaybeAwaitable[A, T], *args, **kwargs
) -> T:  # effectively Awaitable[T]:
    """
    Run the given function asynchronously with the given args. If it is not
    asynchronous, will run in thread. Note: this has to be marked async since in
    some cases we cannot tell ahead of time that `func` is asynchronous so we
    may end up running it to produce a coroutine object which we then need to
    run asynchronously.
    """

    if python_utils.is_really_coroutinefunction(func):
        return await func(*args, **kwargs)

    else:
        res = await asyncio.to_thread(func, *args, **kwargs)

        # HACK010: Might actually have been a coroutine after all.
        if inspect.iscoroutine(res):
            return await res
        else:
            return res


def sync(func: CallableMaybeAwaitable[A, T], *args, **kwargs) -> T:
    """
    Get result of calling function on the given args. If it is awaitable, will
    block until it is finished. Runs in a new thread in such cases.
    """

    if python_utils.is_really_coroutinefunction(func):
        func: Callable[[A], Awaitable[T]]
        awaitable: Awaitable[T] = func(*args, **kwargs)

        # HACK010: Debugging here to make sure it is awaitable.
        assert inspect.isawaitable(awaitable)

        # Check if there is a running loop.
        try:
            loop = asyncio.get_running_loop()

        except Exception:
            # If not, we can create one here and run it until completion.
            loop = asyncio.new_event_loop()
            ret = loop.run_until_complete(awaitable)
            loop.close()
            return ret

        try:
            # If have nest_asyncio, can run in current thread.
            return loop.run_until_complete(awaitable)
        except Exception:
            pass

        # Otherwise we cannot run a new loop in this thread so we create a
        # new thread to run the awaitable until completion.

        def run_in_new_loop():
            th: threading_utils.Thread = current_thread()
            # Attach return value and possibly exception to thread object so we
            # can retrieve from the starter of the thread.
            th.ret = None
            th.error = None
            try:
                loop = asyncio.new_event_loop()
                th.ret = loop.run_until_complete(awaitable)
                loop.close()

            except Exception as e:
                th.error = e

        thread = threading_utils.Thread(target=run_in_new_loop)

        # Start thread and wait for it to finish.
        thread.start()
        thread.join()

        # Get the return or error, return the return or raise the error.
        if thread.error is not None:
            raise thread.error
        else:
            return thread.ret

    else:
        func: Callable[[A], T]
        # Not a coroutine function, so do not need to sync anything.
        # HACK010: TODO: What if the inspect fails here too? We do some checks
        # in desync but not here.

        return func(*args, **kwargs)



"""Serializable app-related classes."""

from __future__ import annotations

from enum import Enum
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)
import warnings

import dill
import pydantic
from trulens.core.feedback import feedback as core_feedback
import trulens.core.instruments as core_instruments
from trulens.core.metric import metric as core_metric
from trulens.core.schema import base as base_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import record as record_schema
from trulens.core.schema import select as select_schema
from trulens.core.schema import types as types_schema
from trulens.core.utils import json as json_utils
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import serial as serial_utils
from trulens.core.utils import text as text_utils
from trulens.core.utils import threading as threading_utils

if TYPE_CHECKING:
    from trulens.core._utils.pycompat import Future  # import style exception
    from trulens.core.database import connector as db_connector
    from trulens.core.feedback import feedback as core_feedback

logger = logging.getLogger(__name__)


class RecordIngestMode(str, Enum):
    """Mode of records ingestion.

    Specify this using the `ingest_mode` to [App][trulens.core.app.App] constructors.
    """

    IMMEDIATE = "immediate"
    """Each record is ingested one by one and written to the database. This is the default mode."""

    BUFFERED = "buffered"
    """Records are buffered and ingested in batches to the database."""


class AppDefinition(pyschema_utils.WithClassInfo, serial_utils.SerialModel):
    """Serialized fields of an app here whereas [App][trulens.core.app.App]
    contains non-serialized fields."""

    app_id: types_schema.AppID = pydantic.Field(frozen=True)  # str
    """Unique identifier for this app.

    Computed deterministically from app_name and app_version. Leaving it here
    for it to be dumped when serializing. Also making it read-only as it should
    not be changed after creation.
    """

    app_name: types_schema.AppName  # str
    """Name for this app. Default is "default_app"."""

    app_version: types_schema.AppVersion  # str
    """Version tag for this app. Default is "base"."""

    tags: types_schema.Tags  # str
    """Tags for the app."""

    metadata: (
        types_schema.Metadata
    )  # dict  # TODO: rename to meta for consistency with other metas
    """Metadata for the app."""

    feedback_definitions: Sequence[types_schema.FeedbackDefinitionID] = []
    """Feedback functions to evaluate on each record."""

    feedback_mode: feedback_schema.FeedbackMode = (
        feedback_schema.FeedbackMode.WITH_APP_THREAD
    )
    """How to evaluate feedback functions upon producing a record."""

    record_ingest_mode: RecordIngestMode
    """Mode of records ingestion."""

    root_class: Optional[pyschema_utils.Class]
    """Class of the main instrumented object.

    Ideally this would be a [ClassVar][typing.ClassVar] but since we want to check this without
    instantiating the subclass of
    [AppDefinition][trulens.core.schema.app.AppDefinition] that would define it, we
    cannot use [ClassVar][typing.ClassVar].
    """

    root_callable: ClassVar[pyschema_utils.FunctionOrMethod]
    """App's main method.

    This is to be filled in by subclass.
    """

    app: serial_utils.JSONized[AppDefinition]
    """Wrapped app in jsonized form."""

    initial_app_loader_dump: Optional[serial_utils.SerialBytes] = None
    """Serialization of a function that loads an app.

    Dump is of the initial app state before any invocations. This can be used to
    create a new session.

    Warning:
        Experimental work in progress.
    """

    app_extra_json: serial_utils.JSON
    """Info to store about the app and to display in dashboard.

    This can be used even if app itself cannot be serialized. `app_extra_json`,
    then, can stand in place for whatever data the user might want to keep track
    of about the app.
    """

    def __init__(
        self,
        app_id: Optional[types_schema.AppName] = None,
        app_name: Optional[types_schema.AppName] = None,
        app_version: Optional[types_schema.AppVersion] = None,
        tags: Optional[types_schema.Tags] = None,
        metadata: Optional[types_schema.Metadata] = None,
        feedback_mode: feedback_schema.FeedbackMode = feedback_schema.FeedbackMode.WITH_APP_THREAD,
        record_ingest_mode: RecordIngestMode = RecordIngestMode.IMMEDIATE,
        app_extra_json: serial_utils.JSON = None,
        **kwargs,
    ):
        kwargs["app_name"] = str(app_name or app_id or "default_app")
        kwargs["app_version"] = app_version or "base"
        kwargs["feedback_mode"] = feedback_mode
        kwargs["tags"] = ""
        kwargs["metadata"] = metadata or {}
        kwargs["app_extra_json"] = app_extra_json or dict()
        kwargs["feedback_definitions"] = [
            f.feedback_definition_id for f in kwargs.get("feedbacks", [])
        ]
        kwargs["record_ingest_mode"] = record_ingest_mode
        kwargs["app_id"] = self._compute_app_id(
            kwargs["app_name"], kwargs["app_version"]
        )
        if app_id is not None and kwargs["app_id"] != app_id:
            # Warning for now:
            warnings.warn(
                "Custom `app_id` values are deprecated. Use `app_name` and/or `app_version` instead.",
                DeprecationWarning,
            )
            # After dep period, change to this:
            # raise ValueError(
            #    "`app_id` does not match with `app_name` and `app_version`! `app_id` is auto-generated and should not need to be passed in!"
            # )

        super().__init__(**kwargs)

        self.record_ingest_mode = record_ingest_mode

        if tags is None:
            tags = "-"  # Set tags to a "-" if None is provided
        self.tags = tags

        # EXPERIMENTAL
        if "initial_app_loader" in kwargs:
            try:
                dump = dill.dumps(kwargs["initial_app_loader"], recurse=True)

                if len(dump) > base_schema.MAX_DILL_SIZE:
                    logger.warning(
                        f"`initial_app_loader` dump is too big ({text_utils.format_quantity(len(dump))}) > {text_utils.format_quantity(base_schema.MAX_DILL_SIZE)} bytes). "
                        "If you are loading large objects, include the loading logic inside `initial_app_loader`.",
                    )
                else:
                    self.initial_app_loader_dump = serial_utils.SerialBytes(
                        data=dump
                    )

                    # This is an older serialization approach that saved things
                    # in local files instead of the DB. Leaving here for now as
                    # serialization of large apps might make this necessary
                    # again.
                    # path_json = Path.cwd() / f"{app_id}.json"
                    # path_dill = Path.cwd() / f"{app_id}.dill"

                    # with path_json.open("w") as fh:
                    #     fh.write(json_utils.json_str_of_obj(self))

                    # with path_dill.open("wb") as fh:
                    #     fh.write(dump)

                    # print(f"Wrote loadable app to {path_json} and {path_dill}.")

            except Exception as e:
                logger.warning(
                    "Could not serialize app loader. "
                    "Some trulens features may not be available: %s",
                    e,
                )

    @staticmethod
    def continue_session(
        app_definition_json: serial_utils.JSON, app: Any
    ) -> AppDefinition:
        # initial_app_loader: Optional[Callable] = None) -> 'AppDefinition':
        """Instantiate the given `app` with the given state
        `app_definition_json`.

        Warning:
            This is an experimental feature with ongoing work.

        Args:
            app_definition_json: The json serialized app.

            app: The app to continue the session with.

        Returns:
            A new `AppDefinition` instance with the given `app` and the given
                `app_definition_json` state.
        """

        app_definition_json["app"] = app

        cls = pyschema_utils.WithClassInfo.get_class(app_definition_json)

        return cls(**app_definition_json)

    @staticmethod
    def _compute_app_id(app_name: str, app_version: str) -> str:
        # Generate a deterministic hash-based ID based on app_name and
        # app_version.
        # NOTE: there is an ongoing discussion about whether to use app_id or
        # directly use/index on the unique combination of
        # (app_name, app_version) when querying the database.
        return json_utils.obj_id_of_obj(
            obj={"app_name": app_name, "app_version": app_version}, prefix="app"
        )

    @staticmethod
    def new_session(
        app_definition_json: serial_utils.JSON,
        initial_app_loader: Optional[Callable] = None,
    ) -> AppDefinition:
        """Create an app instance at the start of a session.

        Warning:
            This is an experimental feature with ongoing work.

        Create a copy of the json serialized app with the enclosed app being
        initialized to its initial state before any records are produced (i.e.
        blank memory).
        """

        serial_bytes_json: Optional[serial_utils.JSON] = app_definition_json[
            "initial_app_loader_dump"
        ]

        if initial_app_loader is None:
            assert (
                serial_bytes_json is not None
            ), "Cannot create new session without `initial_app_loader`."

            serial_bytes = serial_utils.SerialBytes.model_validate(
                serial_bytes_json
            )

            app = dill.loads(serial_bytes.data)()

        else:
            app = initial_app_loader()
            data = dill.dumps(initial_app_loader, recurse=True)
            serial_bytes = serial_utils.SerialBytes(data=data)
            serial_bytes_json = serial_bytes.model_dump()

        app_definition_json["app"] = app
        app_definition_json["initial_app_loader_dump"] = serial_bytes_json

        cls = pyschema_utils.WithClassInfo.get_class(app_definition_json)

        return cls.model_validate_json(app_definition_json)

    def jsonify_extra(self, content):
        # Called by jsonify for us to add any data we might want to add to the
        # serialization of `app`.
        if self.app_extra_json is not None and content["app"] is not None:
            content["app"].update(self.app_extra_json)

        return content

    @staticmethod
    def _submit_feedback_functions(
        record: record_schema.Record,
        feedback_functions: Sequence[core_metric.Metric],
        connector: db_connector.DBConnector,
        app: Optional[AppDefinition] = None,
        on_done: Optional[
            Callable[
                [
                    Union[
                        feedback_schema.FeedbackResult,
                        Future[feedback_schema.FeedbackResult],
                    ]
                ],
                None,
            ]
        ] = None,
    ) -> List[
        Tuple[core_metric.Metric, Future[feedback_schema.FeedbackResult]]
    ]:
        """Schedules to run the given feedback functions.

        Args:
            record: The record on which to evaluate the feedback functions.

            feedback_functions: A collection of feedback functions to evaluate.

            connector: The database connector to use.

            app: The app that produced the given record. If not provided, it is
                looked up from the database of this `TruSession` instance

            on_done: A callback to call when each feedback function is done.

        Returns:

            List[Tuple[Feedback, Future[FeedbackResult]]]

            Produces a list of tuples where the first item in each tuple is the
            feedback function and the second is the future of the feedback result.
        """

        app_id = record.app_id

        if app is None:
            app = AppDefinition.model_validate(connector.get_app(app_id=app_id))
            if app is None:
                raise RuntimeError(
                    f"App {app_id} not present in db. "
                    "Either add it with `TruSession.add_app` or provide `app_json` to `TruSession.run_feedback_functions`."
                )

        else:
            assert (
                app_id == app.app_id
            ), "Record was produced by a different app."

            if connector.get_app(app_id=app.app_id) is None:
                logger.warning(
                    f"App {app_id} was not present in database. Adding it."
                )
                connector.add_app(app=app)

        feedbacks_and_futures = []

        tp = threading_utils.TP()

        for ffunc in feedback_functions:
            # Run feedback function and the on_done callback. This makes sure
            # that Future.result() returns only after on_done has finished.
            def run_and_call_callback(
                ffunc: core_feedback.Feedback,
                app: AppDefinition,
                record: record_schema.Record,
            ):
                token = None
                try:
                    token = core_instruments.do_not_track.set(True)
                    temp = ffunc.run(app=app, record=record)
                    if on_done is not None:
                        try:
                            on_done(temp)
                        finally:
                            return temp
                    return temp
                finally:
                    if token is not None:
                        core_instruments.do_not_track.reset(token)

            fut: Future[feedback_schema.FeedbackResult] = tp.submit(
                run_and_call_callback,
                ffunc=ffunc,
                app=app,
                record=record,
            )

            # Have to roll the on_done callback into the submitted function
            # because the result() is returned before callback runs otherwise.
            # We want to do db work before result is returned.

            feedbacks_and_futures.append((ffunc, fut))

        return feedbacks_and_futures

    @staticmethod
    def get_loadable_apps():
        """Gets a list of all of the loadable apps.

        Warning:
            This is an experimental feature with ongoing work.

        This is those that have `initial_app_loader_dump` set.
        """

        rets = []

        from trulens.core.session import TruSession

        session = TruSession()

        apps = session.get_apps()
        for app in apps:
            dump = app.get("initial_app_loader_dump")
            if dump is not None:
                rets.append(app)

        return rets

    def dict(self):
        # Unsure if the check below is needed. Sometimes we have an `app.App`` but
        # it is considered an `AppDefinition` and is thus using this definition
        # of `dict` instead of the one in `app.App`.
        from trulens.core.app import App

        if isinstance(self, App):
            return json_utils.jsonify(self, instrument=self.instrument)
        else:
            return json_utils.jsonify(self)

    @classmethod
    def select_inputs(cls) -> serial_utils.Lens:
        """Get the path to the main app's call inputs."""

        return select_schema.Select.RecordCall.args

    @classmethod
    def select_outputs(cls) -> serial_utils.Lens:
        """Get the path to the main app's call outputs."""

        return select_schema.Select.RecordCall.rets


# HACK013: Need these if using __future__.annotations .
AppDefinition.model_rebuild()



from concurrent.futures import as_completed
import inspect
import logging
from typing import Optional

from trulens.core.metric import metric as core_metric
from trulens.core.utils import threading as threading_utils

logger = logging.getLogger(__name__)


class context_filter:
    """Provides a decorator to filter contexts based on a given feedback and threshold.

    Args:
        feedback: The feedback object to use for filtering.

        threshold: The minimum feedback value required for a context to be included.

        keyword_for_prompt: Keyword argument to decorator to use for prompt.

    Example:
        ```python
        from trulens.core.guardrails.base import context_filter

        feedback = Feedback(provider.context_relevance, name="Context Relevance")

        class RAG_from_scratch:
            ...
            @context_filter(feedback, 0.5, "query")
            def retrieve(self, *, query: str) -> list:
                results = vector_store.query(
                    query_texts=query,
                    n_results=3
                )
                return [doc for sublist in results['documents'] for doc in sublist]
            ...
        ```
    """

    def __init__(
        self,
        feedback: core_metric.Metric,
        threshold: float,
        keyword_for_prompt: Optional[str] = None,
    ):
        self.feedback = feedback
        self.threshold = threshold
        self.keyword_for_prompt = keyword_for_prompt

    def __call__(self, func):
        sig = inspect.signature(func)

        if self.keyword_for_prompt is not None:
            if self.keyword_for_prompt not in sig.parameters:
                raise TypeError(
                    f"Keyword argument '{self.keyword_for_prompt}' not found in `{func.__name__}` signature."
                )
        else:
            # For backwards compatibility, allow inference of keyword_for_prompt:
            first_arg = list(k for k in sig.parameters.keys() if k != "self")[0]
            self.keyword_for_prompt = first_arg
            logger.warning(
                f"Assuming `{self.keyword_for_prompt}` is the `{func.__name__}` arg to filter. "
                "Specify `keyword_for_prompt` to avoid this warning."
            )

        def wrapper(*args, **kwargs):
            bindings = sig.bind(*args, **kwargs)

            contexts = func(*args, **kwargs)
            with threading_utils.ThreadPoolExecutor(
                max_workers=max(1, len(contexts))
            ) as ex:
                future_to_context = {
                    ex.submit(
                        lambda context=context: self.feedback(
                            bindings.arguments[self.keyword_for_prompt], context
                        )
                    ): context
                    for context in contexts
                }
                filtered = []
                for future in as_completed(future_to_context):
                    context = future_to_context[future]
                    result = future.result()
                    if not isinstance(result, float):
                        raise ValueError(
                            "`context_filter` can only be used with feedback functions that return a float."
                        )
                    if (
                        self.feedback.higher_is_better
                        and result > self.threshold
                    ) or (
                        not self.feedback.higher_is_better
                        and result < self.threshold
                    ):
                        filtered.append(context)
                return filtered

        # note: the following information is manually written to the wrapper because @functools.wraps(func) causes breaking of the method.
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        wrapper.__signature__ = sig
        return wrapper


class block_input:
    """Provides a decorator to block input based on a given feedback and threshold.

    Args:
        feedback: The feedback object to use for blocking.
        threshold: The minimum feedback value required for a context to be included.
        keyword_for_prompt: Keyword argument to decorator to use for prompt.
        return_value: The value to return if the input is blocked. Defaults to None.

    Example:
        ```python
        from trulens.core.guardrails.base import block_input

        feedback = Feedback(provider.criminality, higher_is_better = False)

        class safe_input_chat_app:
            @instrument()
            @block_input(feedback=feedback,
                threshold=0.9,
                keyword_for_prompt="question",
                return_value="I couldn't find an answer to your question.")
            def generate_completion(self, question: str) -> str:
                completion = (
                    oai_client.chat.completions.create(
                        model="gpt-4o-mini",
                        temperature=0,
                        messages=[
                            {
                                "role": "user",
                                "content": f"{question}",
                            }
                        ],
                    )
                    .choices[0]
                    .message.content
                )
                return completion
        ```
    """

    def __init__(
        self,
        feedback: core_metric.Metric,
        threshold: float,
        keyword_for_prompt: Optional[str] = None,
        return_value: Optional[str] = None,
    ):
        self.feedback = feedback
        self.threshold = threshold
        self.keyword_for_prompt = keyword_for_prompt
        self.return_value = return_value

    def __call__(self, func):
        sig = inspect.signature(func)

        if self.keyword_for_prompt is not None:
            if self.keyword_for_prompt not in sig.parameters:
                raise TypeError(
                    f"Keyword argument '{self.keyword_for_prompt}' not found in `{func.__name__}` signature."
                )
        else:
            # For backwards compatibility, allow inference of keyword_for_prompt:
            first_arg = list(k for k in sig.parameters.keys() if k != "self")[0]
            self.keyword_for_prompt = first_arg
            logger.warning(
                f"Assuming `{self.keyword_for_prompt}` is the `{func.__name__}` arg to block on. "
                "Specify `keyword_for_prompt` to avoid this warning."
            )

        def wrapper(*args, **kwargs):
            bindings = sig.bind(*args, **kwargs)
            keyword_value = bindings.arguments[self.keyword_for_prompt]
            result = self.feedback(keyword_value)
            if not isinstance(result, float):
                raise ValueError(
                    "`block_input` can only be used with feedback functions that return a float."
                )
            if (self.feedback.higher_is_better and result < self.threshold) or (
                not self.feedback.higher_is_better and result > self.threshold
            ):
                return self.return_value

            return func(*args, **kwargs)

        # note: the following information is manually written to the wrapper because @functools.wraps(func) causes breaking of the method.
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        wrapper.__signature__ = sig
        return wrapper


class block_output:
    """Provides a decorator to block output based on a given feedback and threshold.

    Args:
        feedback: The feedback object to use for blocking. It must only take a single argument.
        threshold: The minimum feedback value required for a context to be included.
        return_value: The value to return if the input is blocked. Defaults to None.

    Example:
        ```python
        from trulens.core.guardrails.base import block_output

        feedback = Feedback(provider.criminality, higher_is_better = False)

        class safe_output_chat_app:
            @instrument()
            @block_output(feedback = feedback,
                threshold = 0.5,
                return_value = "Sorry, I couldn't find an answer to your question.")
            def chat(self, question: str) -> str:
                completion = (
                    oai_client.chat.completions.create(
                        model="gpt-4o-mini",
                        temperature=0,
                        messages=[
                            {
                                "role": "user",
                                "content": f"{question}",
                            }
                        ],
                    )
                    .choices[0]
                    .message.content
                )
                return completion
        ```
    """

    def __init__(
        self,
        feedback: core_metric.Metric,
        threshold: float,
        return_value: Optional[str] = None,
    ):
        self.feedback = feedback
        self.threshold = threshold
        self.return_value = return_value

    def __call__(self, func):
        sig = inspect.signature(func)

        def wrapper(*args, **kwargs):
            output = func(*args, **kwargs)
            result = self.feedback(output)
            if not isinstance(result, float):
                raise ValueError(
                    "`block_output` can only be used with feedback functions that return a float."
                )
            if (self.feedback.higher_is_better and result < self.threshold) or (
                not self.feedback.higher_is_better and result > self.threshold
            ):
                return self.return_value
            else:
                return output

        # note: the following information is manually written to the wrapper because @functools.wraps(func) causes breaking of the method.
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        wrapper.__signature__ = sig
        return wrapper



from __future__ import annotations

from collections import defaultdict
import contextvars
from dataclasses import dataclass
import functools
import importlib
import inspect
import logging
from pprint import PrettyPrinter
import re
from time import sleep
from types import ModuleType
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)

import pydantic
from pydantic import Field
import requests
from trulens.core.schema import base as base_schema
from trulens.core.utils import asynchro as asynchro_utils
from trulens.core.utils import pace as pace_utils
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import python as python_utils
from trulens.core.utils import serial as serial_utils
from trulens.core.utils import threading as threading_utils

logger = logging.getLogger(__name__)

pp = PrettyPrinter()

A = TypeVar("A")
B = TypeVar("B")
T = TypeVar("T")

INSTRUMENT = "__tru_instrument"

DEFAULT_RPM = 60
"""Default requests per minute for endpoints."""

_NO_CONTEXT_WARNING = """
Cannot find TruLens context. See
https://www.trulens.org/component_guides/other/no_context_warning for more information.
"""

_RE_NO_RETRY = re.compile(
    "("
    + ("|".join(["authentication", "unauthorized", "expired", "quota"]))
    + ")",
    re.IGNORECASE,
)
"""Pattern matched against request exceptions to determine whether they should
be aborted right away instead of retried."""


class EndpointCallback(serial_utils.SerialModel):
    """
    Callbacks to be invoked after various API requests and track various metrics
    like token usage.
    """

    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        arbitrary_types_allowed=True
    )

    endpoint: Endpoint = Field(exclude=True)
    """The endpoint owning this callback."""

    cost: base_schema.Cost = Field(default_factory=base_schema.Cost)
    """Costs tracked by this callback."""

    def handle(self, response: Any) -> None:
        """Called after each request."""
        self.cost.n_requests += 1

    def handle_chunk(self, response: Any) -> None:
        """Called after receiving a chunk from a request."""

        self.cost.n_stream_chunks += 1

    def handle_generation(self, response: Any) -> None:
        """Called after each completion request."""
        self.handle(response)

        self.cost.n_completion_requests += 1

    def handle_generation_chunk(self, response: Any) -> None:
        """Called after receiving a chunk from a completion request."""

        self.handle_chunk(response)

    def handle_classification(self, response: Any) -> None:
        """Called after each classification response."""
        self.handle(response)

        self.cost.n_classification_requests += 1

    def handle_embedding(self, response: Any) -> None:
        """Called after each embedding response."""
        self.handle(response)

        self.cost.n_embedding_requests += 1


class Endpoint(
    pyschema_utils.WithClassInfo,
    serial_utils.SerialModel,
    python_utils.InstanceRefMixin,
):
    """API usage, pacing, and utilities for API endpoints."""

    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        arbitrary_types_allowed=True
    )

    @dataclass
    class EndpointSetup:
        """Class for storing supported endpoint information.

        See [track_all_costs][trulens.core.feedback.endpoint.Endpoint.track_all_costs]
        for usage.
        """

        arg_flag: str
        module_name: str
        class_name: str

    # TODO: factor this out
    BASE_ENDPOINTS: ClassVar[Dict[str, Endpoint]] = {}
    ENDPOINT_SETUPS: ClassVar[List[EndpointSetup]] = [
        EndpointSetup(
            arg_flag="with_openai",
            module_name="trulens.providers.openai.endpoint",
            class_name="OpenAIEndpoint",
        ),
        EndpointSetup(
            arg_flag="with_hugs",
            module_name="trulens.providers.huggingface.endpoint",
            class_name="HuggingfaceEndpoint",
        ),
        EndpointSetup(
            arg_flag="with_litellm",
            module_name="trulens.providers.litellm.endpoint",
            class_name="LiteLLMEndpoint",
        ),
        EndpointSetup(
            arg_flag="with_bedrock",
            module_name="trulens.providers.bedrock.endpoint",
            class_name="BedrockEndpoint",
        ),
        EndpointSetup(
            arg_flag="with_cortex",
            module_name="trulens.providers.cortex.endpoint",
            class_name="CortexEndpoint",
        ),
        EndpointSetup(
            arg_flag="with_dummy",
            module_name="trulens.feedback.dummy.endpoint",
            class_name="DummyEndpoint",
        ),
    ]

    instrumented_methods: ClassVar[
        Dict[Any, List[Tuple[Callable, Callable, Type[Endpoint]]]]
    ] = defaultdict(list)
    """Mapping of classes/module-methods that have been instrumented for cost
     tracking along with the wrapper methods and the class that instrumented
     them.

     Key is the class or module owning the instrumented method. Tuple
     value has:

        - original function,

        - wrapped version,

        - endpoint that did the wrapping.

     """

    name: str
    """API/endpoint name."""

    rpm: float = DEFAULT_RPM
    """Requests per minute."""

    retries: int = 3
    """Retries (if performing requests using this class)."""

    post_headers: Dict[str, str] = Field(default_factory=dict, exclude=True)
    """Optional post headers for post requests if done by this class."""

    pace: pace_utils.Pace = Field(
        default_factory=lambda: pace_utils.Pace(
            marks_per_second=DEFAULT_RPM / 60.0, seconds_per_period=60.0
        ),
        exclude=True,
    )
    """Pacing instance to maintain a desired rpm."""

    global_callback: EndpointCallback = Field(
        exclude=True
    )  # of type _callback_class
    """Track costs not run inside "track_cost" here.

    Also note that Endpoints are singletons (one for each unique name argument)
    hence this global callback will track all requests for the named api even if
    you try to create multiple endpoints (with the same name).
    """

    callback_class: Type[EndpointCallback] = Field(exclude=True)
    """Callback class to use for usage tracking."""

    callback_name: str = Field(exclude=True)
    """Name of variable that stores the callback noted above."""

    _context_endpoints: ClassVar[contextvars.ContextVar] = (
        contextvars.ContextVar("endpoints", default={})
    )
    _context_endpoints.set({})

    def __str__(self):
        # Have to override str/repr due to pydantic issue with recursive models.
        return f"Endpoint({self.name})"

    def __repr__(self):
        # Have to override str/repr due to pydantic issue with recursive models.
        return f"Endpoint({self.name})"

    def __init__(
        self,
        *args,
        name: Optional[str] = None,
        rpm: Optional[float] = None,
        callback_class: Optional[Any] = None,
        _register_instance: bool = True,
        **kwargs,
    ):
        python_utils.InstanceRefMixin.__init__(
            self, register_instance=_register_instance
        )

        if callback_class is None:
            # Some old databases do not have this serialized so lets set it to
            # the parent of callbacks and hope it never gets used.
            callback_class = EndpointCallback
            # raise ValueError(
            #    "Endpoint has to be extended by class that can set `callback_class`."
            # )

        if rpm is None:
            rpm = DEFAULT_RPM
        if name is None:
            name = self.__class__.__name__

        kwargs["name"] = name
        kwargs["callback_class"] = callback_class
        kwargs["global_callback"] = callback_class(endpoint=self)
        kwargs["callback_name"] = f"callback_{name}"
        kwargs["pace"] = pace_utils.Pace(
            seconds_per_period=60.0,  # 1 minute
            marks_per_second=rpm / 60.0,
        )

        super().__init__(*args, **kwargs)

        logger.debug("Creating new endpoint singleton with name %s.", self.name)

        # Extending class should call _instrument_module on the appropriate
        # modules and methods names.

    def pace_me(self) -> float:
        """
        Block until we can make a request to this endpoint to keep pace with
        maximum rpm. Returns time in seconds since last call to this method
        returned.
        """

        return self.pace.mark()

    async def apace_me(self) -> float:
        yield self.pace.amark()

    def _can_retry(self, e: Exception) -> bool:
        """Determine whether a request that raised the given exception can be
        retried.

        Things like authorization errors should not be retried.
        """

        if _RE_NO_RETRY.search(str(e)) is not None:
            return False

        return True

    def run_in_pace(self, func: Callable[[A], B], *args, **kwargs) -> B:
        """Run the given `func` on the given `args` and `kwargs` at pace with the
        endpoint-specified rpm. Failures will be retried `self.retries` times."""

        retries = self.retries + 1
        attempts = 0
        retry_delay = 2.0

        errors = []

        while retries > 0:
            try:
                self.pace_me()
                attempts += 1
                ret = func(*args, **kwargs)
                return ret

            except Exception as e:
                retries -= 1
                logger.error(
                    "%s request failed %s=%s. Retries remaining=%s.",
                    self.name,
                    type(e),
                    e,
                    retries,
                )
                errors.append(e)
                if not self._can_retry(e):
                    break

                if retries > 0:
                    sleep(retry_delay)
                    retry_delay *= 2

        raise RuntimeError(
            f"Endpoint {self.name} request failed {attempts} time(s): \n\t"
            + ("\n\t".join(map(str, errors)))
        )

    def run_me(self, thunk: python_utils.Thunk[T]) -> T:
        """
        DEPRECATED: Run the given thunk, returning itse output, on pace with the api.
        Retries request multiple times if self.retries > 0.

        DEPRECATED: Use `run_in_pace` instead.
        """

        raise NotImplementedError(
            "This method is deprecated. Use `run_in_pace` instead."
        )

    def _instrument_module(self, mod: ModuleType, method_name: str) -> None:
        if python_utils.safe_hasattr(mod, method_name):
            logger.debug(
                "Instrumenting %s.%s for %s",
                python_utils.module_name(mod),
                method_name,
                self.name,
            )
            func = getattr(mod, method_name)
            w = self.wrap_function(func)

            setattr(mod, method_name, w)

            Endpoint.instrumented_methods[mod].append((func, w, type(self)))

    def _instrument_class(self, cls, method_name: str) -> None:
        # Skip Enum classes - they don't allow attribute reassignment
        from enum import EnumMeta

        if isinstance(cls, EnumMeta):
            logger.debug(
                "Skipping instrumentation of enum class %s",
                python_utils.class_name(cls),
            )
            return

        if python_utils.safe_hasattr(cls, method_name):
            func = getattr(cls, method_name)

            # Only instrument if it's actually callable (not an enum value or other attribute)
            if not callable(func):
                logger.debug(
                    "Skipping instrumentation of %s.%s for %s: not callable",
                    python_utils.class_name(cls),
                    method_name,
                    self.name,
                )
                return

            logger.debug(
                "Instrumenting %s.%s for %s",
                python_utils.class_name(cls),
                method_name,
                self.name,
            )
            w = self.wrap_function(func)

            setattr(cls, method_name, w)

            Endpoint.instrumented_methods[cls].append((func, w, type(self)))

    @classmethod
    def print_instrumented(cls):
        """
        Print out all of the methods that have been instrumented for cost
        tracking. This is organized by the classes/modules containing them.
        """

        for wrapped_thing, wrappers in cls.instrumented_methods.items():
            print(
                wrapped_thing
                if wrapped_thing is not object
                else "unknown dynamically generated class(es)"
            )
            for original, _, endpoint in wrappers:
                print(
                    f"\t`{original.__name__}` instrumented "
                    f"by {endpoint} at 0x{id(endpoint):x}"
                )

    def _instrument_class_wrapper(
        self,
        cls,
        wrapper_method_name: str,
        wrapped_method_filter: Callable[[Callable], bool],
    ) -> None:
        """
        Instrument a method `wrapper_method_name` which produces a method so
        that the produced method gets instrumented. Only instruments the
        produced methods if they are matched by named `wrapped_method_filter`.
        """
        if python_utils.safe_hasattr(cls, wrapper_method_name):
            logger.debug(
                "Instrumenting method creator %s.%s for %s",
                cls.__name__,
                wrapper_method_name,
                self.name,
            )
            func = getattr(cls, wrapper_method_name)

            def metawrap(*args, **kwargs):
                produced_func = func(*args, **kwargs)

                if wrapped_method_filter(produced_func):
                    logger.debug(
                        "Instrumenting %s",
                        python_utils.callable_name(produced_func),
                    )

                    instrumented_produced_func = self.wrap_function(
                        produced_func
                    )
                    Endpoint.instrumented_methods[object].append((
                        produced_func,
                        instrumented_produced_func,
                        type(self),
                    ))
                    return instrumented_produced_func
                else:
                    return produced_func

            Endpoint.instrumented_methods[cls].append((
                func,
                metawrap,
                type(self),
            ))

            setattr(cls, wrapper_method_name, metawrap)

    def _instrument_module_members(self, mod: ModuleType, method_name: str):
        if not python_utils.safe_hasattr(mod, INSTRUMENT):
            setattr(mod, INSTRUMENT, set())

        already_instrumented = python_utils.safe_getattr(mod, INSTRUMENT)

        if method_name in already_instrumented:
            logger.debug(
                "module %s already instrumented for %s", mod, method_name
            )
            return

        for m in dir(mod):
            logger.debug(
                "instrumenting module %s member %s for method %s",
                mod,
                m,
                method_name,
            )
            obj = python_utils.safer_getattr(mod, m)
            if obj is not None and isinstance(obj, type):
                # Instrument only classes, not instances.
                self._instrument_class(obj, method_name=method_name)

        already_instrumented.add(method_name)

    @staticmethod
    def track_all_costs(
        __func: asynchro_utils.CallableMaybeAwaitable[A, T],
        *args,
        with_openai: bool = True,
        with_hugs: bool = True,
        with_litellm: bool = True,
        with_bedrock: bool = True,
        with_cortex: bool = True,
        with_dummy: bool = True,
        **kwargs,
    ) -> Tuple[T, Sequence[EndpointCallback]]:
        """Track costs of all of the apis we can currently track, over the
        execution of thunk."""

        endpoints = []

        for endpoint_setup in Endpoint.ENDPOINT_SETUPS:
            if locals().get(endpoint_setup.arg_flag):
                try:
                    mod = importlib.import_module(endpoint_setup.module_name)
                    cls: Type[Endpoint] = python_utils.safe_getattr(
                        mod, endpoint_setup.class_name
                    )
                except ImportError:
                    # If endpoint uses optional packages, will get either module
                    # not found error, or we will have a dummy which will fail
                    # at getattr. Skip either way.
                    continue
                except Exception as e:
                    logger.debug(
                        "Could not import tracking module %s. "
                        "trulens will not track costs/usage of this endpoint. %s",
                        endpoint_setup.module_name,
                        e,
                    )
                    continue

                try:
                    endpoint = next(iter(cls.get_instances()))
                except StopIteration:
                    endpoint = None

                try:
                    if endpoint is None:
                        if cls.__name__ not in Endpoint.BASE_ENDPOINTS:
                            logger.debug(
                                "Could not find an instance of %s. "
                                "trulens will create an endpoint for cost tracking.",
                                cls.__name__,
                            )
                            Endpoint.BASE_ENDPOINTS[cls.__name__] = cls(
                                _register_instance=False
                            )
                        endpoint = Endpoint.BASE_ENDPOINTS[cls.__name__]

                    endpoints.append(endpoint)
                except Exception as e:
                    logger.debug(
                        "Could not initialize endpoint %s. "
                        "Possibly missing key(s). "
                        "trulens will not track costs/usage of this endpoint. %s",
                        cls.__name__,
                        e,
                    )

        return Endpoint._track_costs(
            __func, *args, with_endpoints=endpoints, **kwargs
        )

    @staticmethod
    def track_all_costs_tally(
        __func: asynchro_utils.CallableMaybeAwaitable[A, T],
        *args,
        with_openai: bool = True,
        with_hugs: bool = True,
        with_litellm: bool = True,
        with_bedrock: bool = True,
        with_cortex: bool = True,
        with_dummy: bool = True,
        **kwargs,
    ) -> Tuple[T, python_utils.Thunk[base_schema.Cost]]:
        """Track costs of all of the apis we can currently track, over the
        execution of thunk.

        Returns:
            T: Result of evaluating the thunk.

            Thunk[Cost]: A thunk that returns the total cost of all
                callbacks that tracked costs. This is a thunk as the costs might
                change after this method returns in case of Awaitable results.
        """

        result, cbs = Endpoint.track_all_costs(
            __func,
            *args,
            with_openai=with_openai,
            with_hugs=with_hugs,
            with_litellm=with_litellm,
            with_bedrock=with_bedrock,
            with_cortex=with_cortex,
            with_dummy=with_dummy,
            **kwargs,
        )

        if len(cbs) == 0:
            # Otherwise sum returns "0" below.
            tally = lambda: base_schema.Cost()
        else:
            tally = lambda: sum(cb.cost for cb in cbs)

        return result, tally

    @staticmethod
    def _track_costs(
        __func: asynchro_utils.CallableMaybeAwaitable[A, T],
        *args,
        with_endpoints: Optional[List[Endpoint]] = None,
        **kwargs,
    ) -> Tuple[T, Sequence[EndpointCallback]]:
        """Root of all cost tracking methods.

        Runs the given `thunk`, tracking costs using each of the provided
        endpoints' callbacks.
        """

        if (
            with_endpoints is not None
            and len(with_endpoints) > 0
            and not Endpoint._have_context()
        ):
            # If we cannot access the context vars, we cannot track costs, the
            # last condition issues a warning with more info.
            return __func(*args, **kwargs), []

        # Check to see if this call is within another _track_costs call:
        endpoints = dict(Endpoint._context_endpoints.get())  # copy

        # Collect any new endpoints requested of us.
        with_endpoints = with_endpoints or []

        # Keep track of the new callback objects we create here for returning
        # later.
        callbacks = []

        # Create the callbacks for the new requested endpoints only. Existing
        # endpoints from other frames will keep their callbacks.
        for endpoint in with_endpoints:
            callback_class = endpoint.callback_class
            callback = callback_class(endpoint=endpoint)

            if callback_class not in endpoints:
                endpoints[callback_class] = []

            # And add them to the endpoints dict.
            endpoints[callback_class].append((endpoint, callback))

            callbacks.append(callback)

        # Push the endpoints into the contextvars for wrappers inside the
        # following call to retrieve.
        endpoints_token = Endpoint._context_endpoints.set(endpoints)  # noqa: F841

        try:
            # context_vars = contextvars.copy_context()
            context_vars = {
                Endpoint._context_endpoints: Endpoint._context_endpoints.get()
            }

            # Call the function.
            result: T = __func(*args, **kwargs)

            def rewrap(result):
                if python_utils.is_lazy(result):
                    return python_utils.wrap_lazy(
                        result,
                        wrap=rewrap,
                        context_vars=context_vars,
                    )

                return result

            result = rewrap(result)
        finally:
            # Pop the endpoints from the contextvars.
            # Optionally disable to debug context issues. See App._set_context_vars.
            Endpoint._context_endpoints.reset(endpoints_token)

        # Return result and only the callbacks created here. Outer thunks might
        # return others.
        return result, callbacks

    def track_cost(
        self,
        __func: asynchro_utils.CallableMaybeAwaitable[..., T],
        *args,
        **kwargs,
    ) -> Tuple[T, EndpointCallback]:
        """Tally only the usage performed within the execution of the given
        thunk.

        Returns the thunk's result alongside the EndpointCallback object that
        includes the usage information.
        """

        result, callbacks = Endpoint._track_costs(
            __func, *args, with_endpoints=[self], **kwargs
        )

        return result, callbacks[0]

    def handle_wrapped_call(
        self,
        func: Callable,
        bindings: inspect.BoundArguments,
        response: Any,
        callback: Optional[EndpointCallback],
    ) -> Any:
        """This gets called with the results of every instrumented method.

        This should be implemented by each subclass. Importantly, it must return
        the response or some wrapping of the response.

        Args:
            func: the wrapped method.

            bindings: the inputs to the wrapped method.

            response: whatever the wrapped function returned.

            callback: the callback set up by
                `track_cost` if the wrapped method was called and returned
                within an
                 invocation of `track_cost`.
        """
        return response

    @staticmethod
    def _have_context() -> bool:
        """Determine whether we can access the context vars needed for cost tracking."""

        try:
            Endpoint._context_endpoints.get()

        except LookupError:
            logger.warning(_NO_CONTEXT_WARNING)
            return False

        return True

    def wrap_function(self, func):
        """Create a wrapper of the given function to perform cost tracking."""

        # Skip non-callable objects (e.g., enum values like CallTypes.completion)
        if not callable(func):
            logger.debug(
                "Skipping instrumentation of non-callable %s",
                func,
            )
            return func

        if python_utils.safe_hasattr(func, INSTRUMENT):
            # Store the types of callback classes that will handle calls to the
            # wrapped function in the INSTRUMENT attribute. This will be used to
            # invoke appropriate callbacks when the wrapped function gets
            # called.

            # If INSTRUMENT is set, we don't need to instrument the method again
            # but we may need to add the additional callback class to expected
            # handlers stored at the attribute.

            registered_callback_classes = getattr(func, INSTRUMENT)

            if self.callback_class in registered_callback_classes:
                # If our callback class is already in the list, dont bother
                # adding it again.

                logger.debug(
                    "%s already instrumented for callbacks of type %s",
                    func.__name__,
                    self.callback_class.__name__,
                )

                return func

            else:
                # Otherwise add our callback class but don't instrument again.
                registered_callback_classes += [self.callback_class]
                setattr(func, INSTRUMENT, registered_callback_classes)

                return func

        # If INSTRUMENT is not set, create a wrapper method and return it.
        @functools.wraps(func)
        def tru_wrapper(*args, **kwargs):
            logger.debug(
                "Calling instrumented method %s of type %s, "
                "iscoroutinefunction=%s, "
                "isasyncgeneratorfunction=%s",
                func,
                type(func),
                python_utils.is_really_coroutinefunction(func),
                inspect.isasyncgenfunction(func),
            )

            if not Endpoint._have_context():
                return func(*args, **kwargs)

            endpoints = Endpoint._context_endpoints.get()

            # Context vars as they are now before we reset them. They are
            # captured in the closure of the below.
            context_vars = {
                Endpoint._context_endpoints: Endpoint._context_endpoints.get()
            }

            # Get the result of the wrapped function:
            # response = context_vars.run(func, *args, **kwargs)
            response = func(*args, **kwargs)

            # if len(Endpoint._context_endpoints.get()) == 0:
            #    raise ValueError("No endpoints.")

            bindings = inspect.signature(func).bind(*args, **kwargs)

            # Get all of the callback classes suitable for handling this
            # call. Note that we stored this in the INSTRUMENT attribute of
            # the wrapper method.
            registered_callback_classes = getattr(tru_wrapper, INSTRUMENT)

            # If wrapped method was not called from within _track_costs, we
            # will get None here and do nothing but return wrapped
            # function's response.
            if len(endpoints) == 0:
                logger.debug("No endpoints found.")
                return response

            def update_response(response):
                if python_utils.is_lazy(response):
                    return python_utils.wrap_lazy(
                        response,
                        wrap=update_response,
                        context_vars=context_vars,
                    )

                for callback_class in registered_callback_classes:
                    if callback_class not in endpoints:
                        logger.warning(
                            "Callback class %s is registered for handling %s"
                            " but there are no endpoints waiting to receive the result.",
                            callback_class.__name__,
                            func.__name__,
                        )
                        continue

                    for endpoint, callback in endpoints[callback_class]:
                        logger.debug("Handling endpoint %s.", endpoint.name)

                        with python_utils.with_context(context_vars):
                            response_ = endpoint.handle_wrapped_call(
                                func=func,
                                bindings=bindings,
                                response=response,
                                callback=callback,
                            )
                            if response_ is not None:
                                # Handler is allowed to override the response in
                                # case it wants to wrap some generators or similar
                                # lazy structures.
                                response = response_

                return response

            # Problem here: OpenAI returns generators inside its own special
            # classes. These are thus handled in
            # OpenAIEndpoint.wrapped_call .

            return update_response(response)

        # Set our tracking attribute to tell whether something is already
        # instrumented onto both the sync and async version since either one
        # could be returned from this method.
        setattr(tru_wrapper, INSTRUMENT, [self.callback_class])

        logger.debug("Instrumenting %s for %s.", func.__name__, self.name)

        return tru_wrapper


class _WithPost(Endpoint):
    """Endpoint with post methods."""

    post_headers: Dict[str, str] = Field(default_factory=dict)

    def post(
        self,
        url: str,
        json: serial_utils.JSON,
        timeout: Optional[float] = threading_utils.DEFAULT_NETWORK_TIMEOUT,
    ) -> requests.Response:
        """Make an http post request.

        Subclasses can include additional logic to handle endpoint-specific
        responses.
        """

        self.pace_me()

        return requests.post(
            url, json=json, timeout=timeout, headers=self.post_headers
        )

    async def apost(
        self,
        url: str,
        json: serial_utils.JSON,
        timeout: Optional[float] = threading_utils.DEFAULT_NETWORK_TIMEOUT,
    ) -> requests.Response:
        """Make an http post request.

        Subclasses can include additional logic to handle endpoint-specific
        responses.
        """

        await self.apace_me()

        # TODO: use an asynchronous post method.
        return requests.post(
            url, json=json, timeout=timeout, headers=self.post_headers
        )

    def post_json_first(
        self,
        url: str,
        json: serial_utils.JSON,
        timeout: float = threading_utils.DEFAULT_NETWORK_TIMEOUT,
    ) -> Dict:
        """Wraps `post` with json()[0]."""

        jdata = self.post(url=url, json=json, timeout=timeout).json()
        if len(jdata) == 0:
            raise ValueError("Empty response from post.")

        if len(jdata) > 1:
            logger.warning(
                "Received more than one response from post. "
                "Returning only the first."
            )

        return jdata[0]

    async def apost_json_first(
        self,
        url: str,
        json: serial_utils.JSON,
        timeout: float = threading_utils.DEFAULT_NETWORK_TIMEOUT,
    ) -> Dict:
        """Wraps `apost` with json()[0]."""

        jdata = (await self.apost(url=url, json=json, timeout=timeout)).json()

        if len(jdata) == 0:
            raise ValueError("Empty response from post.")

        if len(jdata) > 1:
            logger.warning(
                "Received more than one response from apost. "
                "Returning only the first."
            )

        return jdata[0]


EndpointCallback.model_rebuild()
Endpoint.model_rebuild()



"""Json utilities and serialization utilities dealing with json."""

from __future__ import annotations

import dataclasses
from enum import Enum
import hashlib
import inspect
import json
import logging
from pathlib import Path
from pprint import PrettyPrinter
import typing
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Optional,
    Sequence,
    Set,
    TypeVar,
    Union,
)

import pydantic
from pydantic.v1 import BaseModel as v1BaseModel
from pydantic.v1.json import ENCODERS_BY_TYPE
from pydantic.v1.json import pydantic_encoder
from trulens.core.utils import constants as constant_utils
from trulens.core.utils import imports as import_utils
from trulens.core.utils import keys as key_utils
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import python as python_utils
from trulens.core.utils import serial as serial_utils

if TYPE_CHECKING:
    from trulens.core import instruments as core_instruments

with import_utils.OptionalImports(
    messages=import_utils.format_import_errors(
        ["openai", "httpx"], purpose="instrumenting apps with OpenAI components"
    )
):
    # httpx.URL and Timeout needed for openai client.
    import httpx
    from openai import Timeout

    def encode_httpx_url(obj: httpx.URL) -> str:
        return str(obj)

    def encode_openai_timeout(obj: Timeout) -> Dict[str, Any]:
        return obj.as_dict()

    ENCODERS_BY_TYPE[httpx.URL] = encode_httpx_url
    ENCODERS_BY_TYPE[Timeout] = encode_openai_timeout

logger = logging.getLogger(__name__)
pp = PrettyPrinter()
T = TypeVar("T")


def _recursive_hash(
    value: Union[dict, list, str, int, bool, float, complex, None],
    ignore_none=False,
) -> str:
    """Hash a json-like structure. Implementation is simplified from merkle_json.

    Args:
        value (Union[dict, list, str, int, bool, float, complex, None]): The value or object to hash.
        ignore_none (bool, optional): If provided, ignore None values in the hash. Defaults to False.

    Returns:
        str: The hash of the value.
    """
    if isinstance(value, list):
        h_acc = [_recursive_hash(v, ignore_none) for v in value]
        return _recursive_hash("".join(sorted(h_acc)), ignore_none)
    elif isinstance(value, dict):
        keys = sorted(value.keys())
        acc = ""
        for k in keys:
            key_val = value[k]
            if ignore_none and key_val is None:
                continue

            acc += f"{k}:{_recursive_hash(key_val, ignore_none=ignore_none)},"
        return _recursive_hash(acc, ignore_none=ignore_none)
    else:
        return hashlib.md5(str(value).encode("utf-8")).hexdigest()


# Add encoders for some types that pydantic cannot handle but we need.


def obj_id_of_obj(obj: Dict[Any, Any], prefix="obj"):
    """
    Create an id from a json-able structure/definition. Should produce the same
    name if definition stays the same.
    """

    return f"{prefix}_hash_{_recursive_hash(obj)}"


def json_str_of_obj(
    obj: Any, *args, redact_keys: bool = False, **kwargs
) -> str:
    """
    Encode the given json object as a string.
    """

    return json.dumps(
        jsonify(obj, *args, redact_keys=redact_keys, **kwargs),
        default=json_default,
        ensure_ascii=False,
    )


def json_default(obj: Any) -> str:
    """
    Produce a representation of an object which does not have a json serializer.
    """

    # Try the encoders included with pydantic first (should handle things like
    # Datetime, and our additional encoders above):
    try:
        return pydantic_encoder(obj)

    except Exception:
        # Otherwise give up and indicate a non-serialization.
        return pyschema_utils.noserio(obj)


def jsonify_for_ui(*args, **kwargs):
    """Options for jsonify common to UI displays.

    Redacts keys and hides special fields introduced by trulens.
    """

    return jsonify(*args, **kwargs, redact_keys=True, skip_specials=True)


def jsonify(
    obj: Any,
    dicted: Optional[Dict[int, serial_utils.JSON]] = None,
    instrument: Optional[core_instruments.Instrument] = None,
    skip_specials: bool = False,
    redact_keys: bool = False,
    include_excluded: bool = True,
    depth: int = 0,
    max_depth: int = 256,
) -> serial_utils.JSON:
    """Convert the given object into types that can be serialized in json.

        Args:
            obj: the object to jsonify.

            dicted: the mapping from addresses of already jsonifed objects (via id)
                to their json.

            instrument: instrumentation functions for checking whether to recur into
                components of `obj`.

            skip_specials: remove specially keyed structures from the json. These
                have keys that start with "__tru_".

            redact_keys: redact secrets from the output. Secrets are detremined by
                `keys.py:redact_value` .

            include_excluded: include fields that are annotated to be excluded by
                pydantic.

            depth: the depth of the serialization of the given object relative to
                the serialization of its container.
    `
            max_depth: the maximum depth of the serialization of the given object.
                Objects to be serialized beyond this will be serialized as
                "non-serialized object" as per `noserio`. Note that this may happen
                for some data layouts like linked lists. This value should be no
                larger than half the value set by
                [sys.setrecursionlimit][sys.setrecursionlimit].

        Returns:
            The jsonified version of the given object. Jsonified means that the the
            object is either a JSON base type, a list, or a dict with the containing
            elements of the same.
    """

    # NOTE(piotrm): We might need to do something special for the below types as
    # they are stateful if iterated. That is, they might be iterable only once
    # and iterating will break their user's interfaces.

    # These are here because we cannot iterate them or await them without
    # breaking the instrumented apps. Instead we return a placeholder value:
    if inspect.isawaitable(obj):
        return "TruLens: Cannot jsonify an awaitable object."
    if isinstance(obj, typing.Iterator):
        return "TruLens: Cannot jsonify an iterator object."
    if inspect.isgenerator(obj):
        return "TruLens: Cannot jsonify a generator object."
    if inspect.isasyncgen(obj):
        return "TruLens: Cannot jsonify an async generator object."

    # These may be necessary to prevent inadvertently stealing some items from
    # the awaitables/generators.
    if inspect.iscoroutine(obj):
        return "TruLens: Cannot jsonify a coroutine object."
    if inspect.isasyncgenfunction(obj):
        return "TruLens: Cannot jsonify an async generator function."
    if inspect.iscoroutinefunction(obj):
        return "TruLens: Cannot jsonify a coroutine function."
    if inspect.isgeneratorfunction(obj):
        return "TruLens: Cannot jsonify a generator function."

    if depth > max_depth:
        logger.debug(
            "Max depth reached for jsonify of object type '%s'.", type(obj)
        )  # careful about str(obj) in case it is recursive infinitely.

        return pyschema_utils.noserio(obj)

    skip_excluded = not include_excluded
    # Hack so that our models do not get exclude dumped which causes many
    # problems. Another variable set here so we can recurse with the original
    # include_excluded .
    if isinstance(obj, serial_utils.SerialModel):
        skip_excluded = True

    from trulens.core.instruments import Instrument

    if instrument is None:
        instrument = Instrument()

    dicted = dicted or {}

    if skip_specials:

        def recur_key(k):
            return (
                isinstance(k, serial_utils.JSON_BASES)
                and k not in constant_utils.ALL_SPECIAL_KEYS
            )

    else:

        def recur_key(k):
            return isinstance(k, serial_utils.JSON_BASES)

    if id(obj) in dicted:
        if skip_specials:
            return None

        return {constant_utils.CIRCLE: id(obj)}

    if isinstance(obj, serial_utils.JSON_BASES):
        if redact_keys and isinstance(obj, str):
            return key_utils.redact_value(obj)

        return obj

    # TODO: remove eventually
    if isinstance(obj, serial_utils.SerialBytes):
        return obj.model_dump()

    if isinstance(obj, Path):
        return str(obj)

    if type(obj) in ENCODERS_BY_TYPE:
        return ENCODERS_BY_TYPE[type(obj)](obj)

    # TODO: should we include duplicates? If so, dicted needs to be adjusted.
    new_dicted = dict(dicted)

    def recur(o):
        return jsonify(
            obj=o,
            dicted=new_dicted,
            instrument=instrument,
            skip_specials=skip_specials,
            redact_keys=redact_keys,
            include_excluded=include_excluded,
            depth=depth + 1,
            max_depth=max_depth,
        )

    content = None

    if isinstance(obj, Enum):
        content = obj.name

    elif isinstance(obj, Dict):
        forward_value = {}
        new_dicted[id(obj)] = forward_value
        forward_value.update({
            k: recur(v) for k, v in obj.items() if recur_key(k)
        })

        # Redact possible secrets based on key name and value.
        if redact_keys:
            for k, v in forward_value.items():
                forward_value[k] = key_utils.redact_value(v=v, k=k)

        content = forward_value

    elif isinstance(obj, Sequence):
        forward_value = []
        new_dicted[id(obj)] = forward_value
        for x in (recur(v) for v in obj):
            forward_value.append(x)

        content = forward_value

    elif isinstance(obj, Set):
        forward_value = []
        new_dicted[id(obj)] = forward_value
        for x in (recur(v) for v in obj):
            forward_value.append(x)

        content = forward_value

    elif isinstance(obj, serial_utils.Lens):  # special handling of paths
        return obj.model_dump()

    elif isinstance(obj, pydantic.BaseModel):
        # Not even trying to use pydantic.dict here.

        forward_value = {}
        new_dicted[id(obj)] = forward_value
        forward_value.update({
            k: recur(python_utils.safe_getattr(obj, k))
            for k, v in type(obj).model_fields.items()
            if (not skip_excluded or not v.exclude) and recur_key(k)
        })

        # Computed fields: read them from the class, not the instance.
        # In Pydantic ≥ 2.10 the public attribute may not be a plain dict.
        # Use the class mapping first, and fall back to __pydantic_computed_fields__.
        cls = type(obj)
        computed_fields_map = getattr(cls, "model_computed_fields", None)
        if not isinstance(computed_fields_map, dict):
            computed_fields_map = getattr(
                cls, "__pydantic_computed_fields__", {}
            )

        for k in computed_fields_map.keys():
            if recur_key(k):
                forward_value[k] = recur(python_utils.safe_getattr(obj, k))

        # Redact possible secrets based on key name and value.
        if redact_keys:
            for k, v in forward_value.items():
                forward_value[k] = key_utils.redact_value(v=v, k=k)

        content = forward_value

    elif isinstance(obj, v1BaseModel):
        # TODO: DEDUP with pydantic.BaseModel case

        # Not even trying to use pydantic.dict here.

        forward_value = {}
        new_dicted[id(obj)] = forward_value
        forward_value.update({
            k: recur(python_utils.safe_getattr(obj, k))
            for k, v in obj.__fields__.items()
            if (not skip_excluded or not v.field_info.exclude) and recur_key(k)
        })

        # Redact possible secrets based on key name and value.
        if redact_keys:
            for k, v in forward_value.items():
                forward_value[k] = key_utils.redact_value(v=v, k=k)

        content = forward_value

    elif dataclasses.is_dataclass(type(obj)):
        # NOTE: cannot use dataclasses.asdict as that may fail due to its use of
        # copy.deepcopy.

        forward_value = {}
        new_dicted[id(obj)] = forward_value

        forward_value.update({
            f.name: recur(python_utils.safe_getattr(obj, f.name))
            for f in dataclasses.fields(obj)
            if recur_key(f.name)
        })

        # Redact possible secrets based on key name and value.
        if redact_keys:
            for k, v in forward_value.items():
                forward_value[k] = key_utils.redact_value(v=v, k=k)

        content = forward_value

    elif instrument.to_instrument_object(obj):
        forward_value = {}
        new_dicted[id(obj)] = forward_value

        kvs = pyschema_utils.clean_attributes(obj, include_props=True)

        # TODO(piotrm): object walks redo
        forward_value.update({
            k: recur(v)
            for k, v in kvs.items()
            if recur_key(k)
            and (
                isinstance(v, serial_utils.JSON_BASES)
                or isinstance(v, Dict)
                or isinstance(v, Sequence)
                or instrument.to_instrument_object(v)
            )
        })

        content = forward_value

    else:
        logger.debug(
            "Do not know how to jsonify an object of type '%s'.", type(obj)
        )  # careful about str(obj) in case it is recursive infinitely.

        content = pyschema_utils.noserio(obj)

    # Add class information for objects that are to be instrumented, known as
    # "components".
    if (
        not skip_specials
        and isinstance(content, dict)
        and not isinstance(obj, dict)
        and (
            instrument.to_instrument_object(obj)
            or isinstance(obj, pyschema_utils.WithClassInfo)
        )
    ):
        content[constant_utils.CLASS_INFO] = pyschema_utils.Class.of_class(
            cls=obj.__class__, with_bases=True
        ).model_dump()

    if not isinstance(obj, serial_utils.Lens) and python_utils.safe_hasattr(
        obj, "jsonify_extra"
    ):
        # Problem with Lens and similar objects: they always say they have every attribute.
        content = obj.jsonify_extra(content)

    return content



# Re-export from feedback.selector for now.
# This ensures Selector is available from the metric module.
# The actual implementation remains in feedback/selector.py to avoid duplication.

from trulens.core.feedback.selector import ProcessedContentNode
from trulens.core.feedback.selector import Selector
from trulens.core.feedback.selector import Trace

__all__ = [
    "Selector",
    "Trace",
    "ProcessedContentNode",
]



"""Utilities for handling deprecation."""

from enum import Enum
import functools
import importlib
import inspect
import logging
from typing import Any, Callable, Dict, Iterable, Optional, Type, Union
import warnings

from trulens.core.utils import imports as import_utils
from trulens.core.utils import python as python_utils

logger = logging.getLogger(__name__)

PACKAGES_MIGRATION_LINK = (
    "https://www.trulens.org/component_guides/other/trulens_eval_migration/"
)


def module_getattr_override(
    module: Optional[str] = None, message: Optional[str] = None
):
    """Override module's `__getattr__` to issue a deprecation errors when
    looking up attributes.

    This expects deprecated names to be prefixed with `DEP_` followed by their
    original pre-deprecation name.

    !!! example
        === "Before deprecation"
            ```python
            # issue module import warning:
            package_dep_warn()

            # define temporary implementations of to-be-deprecated attributes:
            something = ... actual working implementation or alias
            ```

        === "After deprecation"
            ```python
            # define deprecated attribute with None/any value but name with "DEP_"
            # prefix:
            DEP_something = None

            # issue module deprecation warning and override __getattr__ to issue
            # deprecation errors for the above:
            module_getattr_override()
            ```

    Also issues a deprecation warning for the module itself. This will be used
    in the next deprecation stage for throwing errors after deprecation errors.
    """

    if module is None:
        module = python_utils.caller_module_name(offset=1)  # skip our own frame

    mod = importlib.import_module(module)

    # Module deprecation warning.
    packages_dep_warn(module, message=message)

    def _getattr(name: str):
        if not hasattr(mod, "DEP_" + name):
            raise AttributeError(f"module {module} has no attribute {name}")

        # Issue a deprecation warning for the attribute including the
        # optional custom message as well if given.
        raise AttributeError(
            f"Attribute `{name}` has been deprecated in module `{module}`."
            + (" \n" + message if message else "")
        )

    mod.__getattr__ = _getattr


def deprecated_str(s: str, reason: str):
    """Decorator for deprecated string literals."""

    return import_utils.Dummy(
        s, message=reason, original_exception=DeprecationWarning(reason)
    )


def is_deprecated(obj: Any):
    """Check if object is deprecated.

    Presently only supports values created by `deprecated_str`.
    """

    if import_utils.is_dummy(obj):
        ex = inspect.getattr_static(obj, "original_exception")
        if isinstance(ex, DeprecationWarning):
            return True

    return True


def deprecated_property(message: str):
    """Decorator for deprecated attributes defined as properties."""

    warned = False

    def wrapper(func):
        @functools.wraps(func)
        def _movedattribute(*args, **kwargs):
            nonlocal warned
            if not warned:
                warnings.warn(message, DeprecationWarning, stacklevel=2)
                warned = True
            return func(*args, **kwargs)

        return property(_movedattribute)

    return wrapper


def handle_deprecated_kwarg(
    kwargs: Dict[str, Any],
    old_name: str,
    new_name: str,
    new_value: Optional[str],
    stacklevel: int = 3,
) -> Optional[str]:
    """Handle deprecated keyword argument that has been renamed.

    This helper checks if an old parameter name is in kwargs, emits a
    deprecation warning, and returns the appropriate value to use.

    Args:
        kwargs: The kwargs dictionary to check and modify. The old key will
            be removed from this dictionary if present.
        old_name: The deprecated parameter name.
        new_name: The new parameter name.
        new_value: The current value of the new parameter.
        stacklevel: Stack level for the warning. Defaults to 3 (caller's caller).

    Returns:
        The value to use for the new parameter. If new_value is None and the
        old parameter was provided, returns the old parameter's value.
        Otherwise returns new_value unchanged.

    Example:
        ```python
        def my_function(new_param: Optional[str] = None, **kwargs):
            new_param = handle_deprecated_kwarg(
                kwargs, "old_param", "new_param", new_param
            )
        ```
    """
    if old_name in kwargs:
        warnings.warn(
            f"Parameter `{old_name}` has been renamed to `{new_name}`. "
            f"Please update your code to use `{new_name}` instead.",
            DeprecationWarning,
            stacklevel=stacklevel,
        )
        if new_value is None:
            new_value = kwargs.pop(old_name)
        else:
            kwargs.pop(old_name)
    return new_value


def packages_dep_warn(
    module: Optional[str] = None, message: Optional[str] = None
):
    """Issue a deprecation warning for a backwards-compatibility modules.

    This is specifically for the trulens_eval -> trulens module renaming and
    reorganization. If `message` is given, that is included first in the
    deprecation warning.
    """

    if module is None:
        module = python_utils.caller_module_name(offset=1)  # skip our own frame

    full_message = (
        f"The `{module}` module is deprecated. "
        f"See {PACKAGES_MIGRATION_LINK} for instructions on migrating to `trulens.*` modules."
    )

    if message is not None:
        # Put the custom message first.
        full_message += message + "\n\n" + full_message

    warnings.warn(
        full_message,
        DeprecationWarning,
        stacklevel=3,
    )


def has_deprecated(obj: Union[Callable, Type]) -> bool:
    """Check if a function or class has been deprecated."""

    return has_moved(obj)


def has_moved(obj: Union[Callable, Type]) -> bool:
    """Check if a function or class has been moved."""

    return (
        hasattr(obj, "__doc__")
        and obj.__doc__ is not None
        and "has moved:\n" in obj.__doc__
    )


def staticmethod_renamed(new_name: str):
    """Issue a warning upon static method call that has been renamed or moved.

    Issues the warning only once.
    """

    warned = False

    def wrapper(func):
        old_name = python_utils.callable_name(func)

        message = f"Static method `{old_name}` has been renamed or moved to `{new_name}`.\n"

        @functools.wraps(func)
        def _renamedmethod(*args, **kwargs):
            nonlocal warned, old_name

            if not warned:
                warnings.warn(message, DeprecationWarning, stacklevel=2)
                warned = True

            return func(*args, **kwargs)

        _renamedmethod.__doc__ = message

        return _renamedmethod

    return wrapper


def method_renamed(new_name: str):
    """Issue a warning upon method call that has been renamed or moved.

    Issues the warning only once.
    """

    warned = False

    def wrapper(func):
        old_name = func.__name__

        message = (
            f"Method `{old_name}` has been renamed or moved to `{new_name}`.\n"
        )

        @functools.wraps(func)
        def _renamedmethod(self, *args, **kwargs):
            nonlocal warned, old_name

            if not warned:
                warnings.warn(message, DeprecationWarning, stacklevel=2)
                warned = True

            return func(self, *args, **kwargs)

        _renamedmethod.__doc__ = message

        return _renamedmethod

    return wrapper


def function_moved(func: Callable, old: str, new: str):
    """Issue a warning upon function call that has been moved to a new location.

    Issues the warning only once. The given callable must have a name, so it
    cannot be a lambda.
    """

    if not hasattr(func, "__name__"):
        raise TypeError("Cannot create dep message without function name.")

    message = (
        f"Function `{func.__name__}` has moved:\n"
        f"\tOld import: `from {old} import {func.__name__}`\n"
        f"\tNew import: `from {new} import {func.__name__}`\n "
    )

    warned = False

    @functools.wraps(func)
    def _movedfunction(*args, **kwargs):
        nonlocal warned

        if not warned:
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            warned = True

        return func(*args, **kwargs)

    _movedfunction.__doc__ = message

    return _movedfunction


def class_moved(
    cls: Type,
    old_location: Optional[str] = None,
    new_location: Optional[str] = None,
):
    """Issue a warning upon class instantiation that has been moved to a new
    location.

    Issues the warning only once.
    """

    return cls

    # NOTE(piotrm): Temporarily disabling the class moved mechanism as it
    # interferes with Feedback serialization.

    if new_location is None:
        new_location = cls.__module__

    message = (
        f"Class `{cls.__name__}` has moved:\n"
        + (
            f"\tOld import: `from {old_location} import {cls.__name__}`\n"
            if old_location
            else ""
        )
        + f"\tNew import: `from {new_location} import {cls.__name__}`\n "
        + f"See {PACKAGES_MIGRATION_LINK} for instructions on migrating to `trulens` modules."
    )

    warned = False
    moved_class = cls

    if issubclass(cls, Enum):
        # Enums cannot be extended.
        return cls

    class _MovedClass(moved_class):
        def __new__(cls, *args, **kwargs):
            nonlocal warned

            if not warned:
                warnings.warn(message, DeprecationWarning, stacklevel=2)
                warned = True

            if isinstance(moved_class.__new__, classmethod):
                return moved_class.__new__(cls, *args, **kwargs)
            else:
                return moved_class.__new__(cls)

    _MovedClass.__doc__ = message

    return _MovedClass


def moved(
    globals_dict: Dict[str, Any],
    old: Optional[str] = None,
    new: Optional[str] = None,
    names: Optional[Iterable[str]] = None,
):
    """Replace all classes or function in the given dictionary with ones that
    issue a deprecation warning upon initialization or invocation.

    You can use this with module `globals_dict=globals()` and `names=__all__` to
    deprecate all exposed module members.

    Args:
        globals_dict: The dictionary to update. See [globals][globals].

        old: The old location of the classes.

        new: The new location of the classes.

        names: The names of the classes or functions to update. If None, all
            classes and functions are updated.
    """

    if names is None:
        names = globals_dict.keys()

    for name in names:
        val = globals_dict[name]
        if isinstance(val, import_utils.Dummy):
            # skip dummies
            continue

        if inspect.isclass(val):
            globals_dict[name] = class_moved(val, old, new)
        elif inspect.isfunction(val):
            globals_dict[name] = function_moved(val, old, new)
        else:
            logger.warning("I don't know how to move %s", name)



# Metric module - new primary API for evaluation metrics.
# This replaces the feedback module (which is now deprecated).

from trulens.core.metric.metric import InvalidSelector
from trulens.core.metric.metric import Metric
from trulens.core.metric.metric import SkipEval
from trulens.core.metric.selector import Selector

__all__ = [
    "Metric",
    "Selector",
    "SkipEval",
    "InvalidSelector",
]



"""
DEPRECATED: Client-side custom metrics functionality.

This module is deprecated. Use trulens.core.metric.Metric instead.

The MetricConfig class is maintained for backward compatibility but will be
removed in a future version.
"""

from __future__ import annotations

import inspect
from typing import Any, Callable, Dict, Optional
import warnings

from trulens.core.feedback.selector import Selector
from trulens.core.metric.metric import Metric


class MetricConfig:
    """
    DEPRECATED: Use Metric instead.

    This class is maintained for backward compatibility. All functionality
    has been merged into the Metric class.

    Example of migrating to Metric:
        ```python
        # Old way (deprecated):
        from trulens.core.feedback import MetricConfig
        config = MetricConfig(
            metric_name="my_metric",
            metric_implementation=my_fn,
            selectors={"arg": Selector.select_record_input()},
        )
        feedback = config.create_feedback_definition()

        # New way (recommended):
        from trulens.core import Metric, Selector
        metric = Metric(
            name="my_metric",
            implementation=my_fn,
        ).on({"arg": Selector.select_record_input()})
        ```
    """

    def __init__(
        self,
        metric_name: str,
        metric_implementation: Callable,
        metric_type: Optional[str] = None,
        selectors: Optional[Dict[str, Selector]] = None,
        computation_type: str = "client",
        higher_is_better: bool = True,
        description: Optional[str] = None,
    ):
        """
        Initialize a metric configuration (deprecated, use Metric instead).

        Args:
            metric_name: Unique semantic identifier for this specific metric usage
                        (e.g., "text2sql_accuracy_v1", "custom_relevance_for_qa")
            metric_implementation: The metric function to execute
            metric_type: Implementation identifier of the custom metric
                        (e.g., "text2sql", "accuracy", "relevance").
                        If not provided, defaults to the function name.
            selectors: Dictionary mapping parameter names to Selectors
            computation_type: Where to compute ("client" or "server")
            higher_is_better: Whether higher scores are better
            description: Optional description of the metric
        """
        warnings.warn(
            "MetricConfig is deprecated and will be removed in a future version. "
            "Use Metric instead:\n"
            "  from trulens.core import Metric, Selector\n"
            "  metric = Metric(\n"
            "      name='...',\n"
            "      implementation=fn,\n"
            "  ).on({'arg': Selector.select_record_input()})",
            DeprecationWarning,
            stacklevel=2,
        )

        self.metric_name = metric_name
        self.metric_implementation = metric_implementation
        self.metric_type = metric_type or (
            metric_implementation.__name__
            if hasattr(metric_implementation, "__name__")
            else "custom_metric"
        )
        self.computation_type = computation_type
        self.selectors = selectors or {}
        self.higher_is_better = higher_is_better
        self.description = description

        # Validate function signature
        self._validate_function()

    def _validate_function(self) -> None:
        """Validate that the function has a proper signature for a metric."""
        if not callable(self.metric_implementation):
            raise ValueError("metric_implementation must be callable")

        sig = inspect.signature(self.metric_implementation)

        # Check that function has at least one parameter
        if len(sig.parameters) == 0:
            raise ValueError(
                f"Metric function '{self.metric_name}' must have at least one parameter"
            )

    def validate_selectors(self) -> None:
        """
        Validate that selectors match the function signature.

        Only checks required parameters (those without default values).

        Raises:
            ValueError: If selectors don't match function parameters
        """
        sig = inspect.signature(self.metric_implementation)

        # Get all function parameters
        function_args = list(sig.parameters.keys())

        # Get only required parameters (those without defaults)
        # Exclude *args and **kwargs as they are never required selectors
        required_function_args = [
            param_name
            for param_name, param in sig.parameters.items()
            if param.default == inspect.Parameter.empty
            and param.kind
            not in (
                inspect.Parameter.VAR_POSITIONAL,
                inspect.Parameter.VAR_KEYWORD,
            )
        ]

        error_msg = ""

        # Check for extra selectors
        extra_selectors = []
        for selector in self.selectors:
            if selector not in function_args:
                extra_selectors.append(selector)
        if extra_selectors:
            error_msg += f"Metric function '{self.metric_name}' has selectors that are not in the function signature:\n"
            error_msg += f"Extra selectors: {extra_selectors}\n"
            error_msg += f"Function args: {function_args}\n"

        # Check for missing required selectors (only required parameters)
        missing_selectors = []
        for required_function_arg in required_function_args:
            if required_function_arg not in self.selectors:
                missing_selectors.append(required_function_arg)
        if missing_selectors:
            error_msg += (
                f"Metric function '{self.metric_name}' has missing selectors:\n"
            )
            error_msg += f"Missing selectors: {missing_selectors}\n"
            error_msg += f"Required function args: {required_function_args}\n"

        if error_msg:
            raise ValueError(error_msg)

    def create_feedback_definition(self) -> Metric:
        """
        Create a Metric instance from this metric configuration.

        DEPRECATED: This method returns a Metric instance (formerly Feedback).

        Returns:
            A Metric instance configured for this metric

        Raises:
            ValueError: If selectors don't match function parameters
        """
        self.validate_selectors()

        # For client-side metrics, we don't need serialization since they won't be deferred
        # Create a non-serializable implementation placeholder to avoid serialization warnings
        from trulens.core.utils import pyschema as pyschema_utils

        try:
            # Try to create a serializable implementation
            implementation = pyschema_utils.FunctionOrMethod.of_callable(
                self.metric_implementation, loadable=True
            )
        except Exception:
            # If serialization fails (e.g., function in __main__, like client side metrics defined in a notebook), create non-loadable version
            # This is fine for client-side metrics that execute immediately
            implementation = pyschema_utils.FunctionOrMethod.of_callable(
                self.metric_implementation, loadable=False
            )

        return Metric(
            imp=self.metric_implementation,
            implementation=implementation,  # Pre-provide serialized version
            name=self.metric_name,
            higher_is_better=self.higher_is_better,
            metric_type=self.metric_type,
            description=self.description,
        ).on(self.selectors)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the metric config to a dictionary.

        Returns:
            Dictionary representation of the config
        """
        return {
            "metric_name": self.metric_name,
            "metric_type": self.metric_type,
            "computation_type": self.computation_type,
            "selectors": self.selectors,
            "higher_is_better": self.higher_is_better,
            "description": self.description,
        }

    def __repr__(self) -> str:
        return (
            f"MetricConfig(metric_name='{self.metric_name}', "
            f"metric_type='{self.metric_type}', "
            f"computation='{self.computation_type}', selectors={len(self.selectors)})"
        )



from __future__ import annotations

import json
import traceback
from typing import Callable, Dict, List, Optional, Tuple

from sqlalchemy import select
from sqlalchemy.orm import Session
from trulens.core.database import migrations as db_migrations
from trulens.core.database.base import DB
from trulens.core.database.legacy import migration as legacy_migration
from trulens.core.schema import app as app_schema
from trulens.core.schema import base as base_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import record as record_schema
from trulens.core.utils import pyschema as pyschema_utils

sql_alchemy_migration_versions: List[int] = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
"""DB versions."""

sqlalchemy_upgrade_paths: Dict[int, Tuple[int, Callable[[DB]]]] = {
    # Dict Structure:
    # from_version: (to_version, migrate_method)
    # Example:
    # 1: (2, migrate_alembic_1_to_2)
}
"""A DAG of upgrade functions to get to most recent DB."""


def _get_sql_alchemy_compatibility_version(version: int) -> int:
    """Gets the last compatible version of a DB that needed data migration.

    Args:
        version: The alembic version

    Returns:
        int: An alembic version of the oldest compatible DB
    """

    for candidate_version in sorted(
        sql_alchemy_migration_versions, reverse=True
    ):
        if candidate_version <= version:
            return candidate_version
    raise ValueError(
        f"The version given {version} is not compatible with any known version! Known versions: {sql_alchemy_migration_versions}"
    )


def _sql_alchemy_serialization_asserts(db: DB) -> None:
    """Checks that data migrated JSONs can be deserialized from DB to Python objects.

    Args:
        db (DB): The database object

    Raises:
        VersionException: raises if a serialization fails
    """
    session = Session(db.engine)

    import inspect

    # from trulens.core.database import orm
    # Dynamically check the orm classes since these could change version to version
    for _, orm_obj in inspect.getmembers(db.orm):
        # Check only classes
        if inspect.isclass(orm_obj):
            mod_check = str(orm_obj).split(".")

            # Check only orm defined classes
            if (
                len(mod_check) > 2 and "orm" == mod_check[-2]
            ):  # <class mod.mod.mod.orm.SQLORM>
                stmt = select(orm_obj)

                # for each record in this orm table
                for db_record in session.scalars(stmt).all():
                    # for each column in the record
                    for attr_name in db_record.__dict__:
                        # Check only json columns
                        if "_json" in attr_name:
                            db_json_str = getattr(db_record, attr_name)
                            if (
                                db_json_str
                                == legacy_migration.MIGRATION_UNKNOWN_STR
                            ):
                                continue

                            # Do not check Nullables
                            if db_json_str is not None:
                                # Test deserialization
                                test_json = json.loads(
                                    getattr(db_record, attr_name)
                                )

                                # special implementation checks for serialized classes
                                if "implementation" in test_json:
                                    try:
                                        pyschema_utils.FunctionOrMethod.model_validate(
                                            test_json["implementation"]
                                        ).load()
                                    except ImportError:
                                        # Import error is not a migration problem.
                                        # It signals that the function cannot be used for deferred evaluation.
                                        pass

                                if attr_name == "record_json":
                                    record_schema.Record.model_validate(
                                        test_json
                                    )
                                elif attr_name == "cost_json":
                                    base_schema.Cost.model_validate(test_json)
                                elif attr_name == "perf_json":
                                    base_schema.Perf.model_validate(test_json)
                                elif attr_name == "calls_json":
                                    for record_app_call_json in test_json[
                                        "calls"
                                    ]:
                                        feedback_schema.FeedbackCall.model_validate(
                                            record_app_call_json
                                        )
                                elif attr_name == "feedback_json":
                                    feedback_schema.FeedbackDefinition.model_validate(
                                        test_json
                                    )
                                elif attr_name == "app_json":
                                    app_schema.AppDefinition.model_validate(
                                        test_json
                                    )
                                else:
                                    # If this happens, trulens needs to add a migration
                                    raise legacy_migration.VersionException(
                                        f"serialized column migration not implemented: {attr_name}."
                                    )


def data_migrate(db: DB, from_version: Optional[str]):
    """Makes any data changes needed for upgrading from the from_version to the
    current version.

    Args:
        db: The database instance.

        from_version: The version to migrate data from.

    Raises:
        VersionException: Can raise a migration or validation upgrade error.
    """

    if from_version is None:
        sql_alchemy_from_version = 1
    else:
        sql_alchemy_from_version = int(from_version)
    from_compat_version = _get_sql_alchemy_compatibility_version(
        sql_alchemy_from_version
    )
    to_compat_version = None
    fail_advice = "Please open a ticket on trulens github page including this error message. The migration completed so you can still proceed; but stability is not guaranteed. If needed, you can `TruSession.reset_database()`"

    try:
        while from_compat_version in sqlalchemy_upgrade_paths:
            to_compat_version, migrate_fn = sqlalchemy_upgrade_paths[
                from_compat_version
            ]
            migrate_fn(db)
            from_compat_version = to_compat_version
        print("DB Migration complete!")
    except Exception as e:
        tb = traceback.format_exc()
        current_revision = db_migrations.DbRevisions.load(db.engine).current
        raise legacy_migration.VersionException(
            f"Migration failed on {db} from db version - {from_version} on step: {str(to_compat_version)}. The attempted DB version is {current_revision} \n\n{tb}\n\n{fail_advice}"
        ) from e
    try:
        _sql_alchemy_serialization_asserts(db)
        print("DB Validation complete!")
    except Exception as e:
        tb = traceback.format_exc()
        current_revision = db_migrations.DbRevisions.load(db.engine).current
        raise legacy_migration.VersionException(
            f"Validation failed on {db} from db version - {from_version} on step: {str(to_compat_version)}. The attempted DB version is {current_revision} \n\n{tb}\n\n{fail_advice}"
        ) from e



"""Serializable dataset-related classes."""

from __future__ import annotations

import logging
from typing import Hashable, Optional

from trulens.core.schema import types as types_schema
from trulens.core.utils import json as json_utils
from trulens.core.utils import serial as serial_utils

logger = logging.getLogger(__name__)


class Dataset(serial_utils.SerialModel, Hashable):
    """The class that holds the metadata of a dataset stored in the DB."""

    dataset_id: types_schema.DatasetID  # str
    """The unique identifier for the dataset."""

    name: str
    """The name of the dataset."""

    meta: types_schema.Metadata  # dict
    """Metadata associated with the dataset."""

    def __init__(
        self,
        name: str,
        dataset_id: Optional[types_schema.DatasetID] = None,
        meta: Optional[types_schema.Metadata] = None,
        **kwargs,
    ):
        kwargs["name"] = name
        kwargs["meta"] = meta if meta is not None else {}
        super().__init__(
            dataset_id="temporary", **kwargs
        )  # dataset_id will be updated below

        if dataset_id is None:
            dataset_id = json_utils.obj_id_of_obj(
                json_utils.jsonify(self), prefix="dataset"
            )

        self.dataset_id = dataset_id

    def __hash__(self):
        return hash(self.dataset_id)


# HACK013: Need these if using __future__.annotations .
Dataset.model_rebuild()
